/**
 * Copyright (c) 2022
 *    Marcus Britanicus (https://gitlab.com/marcusbritanicus)
 *    Abrar (https://gitlab.com/s96Abrar)
 *    rahmanshaber (https://gitlab.com/rahmanshaber)
 *
 * The DFL::Settings class extends the QSettings class.
 **/

#include <QGuiApplication>
#include <QFontDatabase>
#include <QLibraryInfo>
#include <QScreen>
#include <QDebug>
#include <QFile>

#include "Settings.hpp"
#include "SettingsImpl.hpp"

DFL::Settings::Settings( QString project, QString app, QString ConfigPath ) {
    /** Initialize our implementation class */
    impl = new DFL::SettingsImpl( project, app, ConfigPath );

    /** XDG Config Path: This is baked in when compiling Qt. Typically, /etc/xdg */
    QString xdgConfigPath = QLibraryInfo::location( QLibraryInfo::SettingsPath );

    QFileSystemWatcher *fsw = new QFileSystemWatcher( this );

    fsw->addPath( impl->defSett->fileName() );
    fsw->addPath( impl->sysSett->fileName() );
    fsw->addPath( impl->usrSett->fileName() );

    connect(
        fsw, &QFileSystemWatcher::fileChanged, [ = ]( const QString& file ) {
            /** Get the changed settings and emit the changes */
            emitSettings();

            /* File deleted and created again: add it to the watcher */
            if ( not fsw->files().contains( file ) and QFile::exists( file ) ) {
                fsw->addPath( file );
            }
        }
    );
}


DFL::Settings::Proxy DFL::Settings::value( const QString& key ) {
    QVariant rawVal( impl->value( key ) );

    return Proxy{ rawVal };
}


void DFL::Settings::setValue( const QString& key, QVariant value ) {
    if ( not impl->defSett->contains( key ) and not key.startsWith( "Session/" ) ) {
        qDebug() << "Unknown key:" << key;
        qDebug() << "Report this incident to the developers.";

        return;
    }

    impl->setValue( key, value );
}


void DFL::Settings::emitSettings() {
    /* We are bothered only about changes to user and system settings */
    ChangedSettings chSettList = impl->changedSettings();

    bool defaultChanged = false;

    for ( ChangedSetting s: chSettList ) {
        if ( s.state == 'm' ) {
            emit settingChanged( s.key, s.value );
        }

        else if ( (s.state == 'a') or (s.state == 'd') ) {
            defaultChanged = true;
        }
    }

    if ( defaultChanged ) {
        emit restartApp();
    }
}


QStringList DFL::Settings::allKeys() const {
    return impl->allKeys();
}


bool DFL::Settings::hasKey( QString key ) const {
    return impl->hasKey( key );
}
