/**
 * Copyright (c) 2022
 *    Marcus Britanicus (https://gitlab.com/marcusbritanicus)
 *    Abrar (https://gitlab.com/s96Abrar)
 *    rahmanshaber (https://gitlab.com/rahmanshaber)
 *
 * The DFL::Settings expands the QSettings class.
 *
 * IMPL Class
 *  This class will read the various settings files
 *  The class structure and working is inspired by QSettings.
 *  This is purely an implementational detail.
 *  This class can be modified/removed at anytime.
 **/

#include <cmath>

#include <QFontDatabase>
#include <QStandardPaths>
#include <QGuiApplication>
#include <QLibraryInfo>
#include <QProcess>
#include <QScreen>
#include <QDir>

#include <QDebug>

#if QT_VERSION < QT_VERSION_CHECK( 6, 0, 0 )
#include <QTouchDevice>
#endif

#include "SettingsImpl.hpp"

static inline bool isTouchCapable() {
    QProcess proc;

    proc.start( "udevadm", QStringList() << "info" << "--export-db" );
    proc.waitForFinished( -1 );

    QString output = QString::fromLocal8Bit( proc.readAllStandardOutput() + '\n' + proc.readAllStandardError() );

    if ( output.contains( "ID_INPUT_TOUCHSCREEN=1" ) ) {
        return true;
    }

    return false;
}


static inline int getFormFactor() {
    /**
     * Form Factors
     * ------------
     *
     * 0: Desktop - Typically 14" to 16" screens
     * 1: Tablets - Typically  7" to 10" screens
     * 2: Mobiles - Typically  4" to  6" screens
     * 3: Monitor - Typically larger than 16" - huge sizes, we're not built for these screens.
     **/

    QSizeF screenSize = QGuiApplication::primaryScreen()->physicalSize();
    double diag       = sqrt( pow( screenSize.width(), 2 ) + pow( screenSize.height(), 2 ) ) / 25.4;

    //- Deady large screen mobiles
    if ( diag <= 6.5 ) {
        return 2;
    }

    //- Tablets and small netbooks
    else if ( diag <= 10.1 ) {
        return 1;
    }

    //- Typical desktops
    else if ( diag <= 16.0 ) {
        return 0;
    }

    // Large displays
    else {
        return 3;
    }

    /* Default value */
    return 0;
}


DFL::SettingsImpl::SettingsImpl( QString project, QString app, QString ConfigPath ) {
    /** XDG Config Path: This is baked in when compiling Qt. Typically, /etc/xdg */
    QString xdgConfigPath = QLibraryInfo::location( QLibraryInfo::SettingsPath );

    /** Config path of the user: we want the generic location. */
    QString xdgHomeConfigPath = QStandardPaths::writableLocation( QStandardPaths::GenericConfigLocation );

    defSett = new QSettings( ConfigPath + "/" + app + ".conf", QSettings::IniFormat );
    sysSett = new QSettings( xdgConfigPath + "/" + project + "/" + app + ".conf", QSettings::IniFormat );
    usrSett = new QSettings( xdgHomeConfigPath + "/" + project + "/" + app + ".conf", QSettings::IniFormat );

    /** Populate the settings map */
    map = readCurrentSettings();
}


QVariant DFL::SettingsImpl::value( const QString& key ) {
    /**
     * SafetyNet: If @map does not this key, then it's probably an invalid key.
     * We will simply return a invalid QVariant(). The app should handle this
     * suitably.
     */
    if ( not map.contains( key ) ) {
        return QVariant();
    }

    /** Get the form factor */
    if ( key == "FormFactor" ) {
        return getFormFactor();
    }

    /**
     * Some parsing may be needed at this point
     * We convert environment variables, shell short-forms etc, here.
     */
    QVariant value = map.value( key );

    /* TouchMode: As seen by the user */
    if ( key == "TouchMode" ) {
        int touch = value.toInt();
        switch ( touch ) {
            /* Auto detect */
            case 0: {
                bool yes = false;
#if (QT_VERSION >= QT_VERSION_CHECK( 6, 0, 0 ) )
                    yes = isTouchCapable();

#else
                    QList<const QTouchDevice *> devs = QTouchDevice::devices();
                    for ( const QTouchDevice *dev: devs ) {
                        if ( dev->type() == QTouchDevice::TouchScreen ) {
                            yes = true;
                        }
                    }
#endif

                return QVariant( yes );
            }

            /* TouchMode is on */
            case 1: {
                return QVariant( true );
            }

            /* TouchMode is off */
            case 2: {
                return QVariant( false );
            }

            default: {
                return QVariant( false );
            }
        }
    }

    /* TouchModeRaw: Raw value saved in the file */
    else if ( key == "TouchModeRaw" ) {
        return value;
    }

    /* Get the default value for the key */
    else {
        if ( (QMetaType::Type)value.type() == QMetaType::QString ) {
            QString strValue = value.toString();

            /** Replace ~ with home path, but only if the string begins with ~ */
            if ( strValue.startsWith( "~/" ) or strValue == "~" ) {
                strValue = strValue.replace( "~", QDir::homePath() );

                return QVariant( strValue );
            }

            /** Envisonment variable */
            else if ( strValue.startsWith( "$" ) ) {
                QString envVar      = strValue.split( "/" ).takeFirst();
                QString envVarValue = qEnvironmentVariable( envVar.toUtf8().constData() );

                if ( not envVarValue.isEmpty() ) {
                    strValue = strValue.replace( envVar, envVarValue );
                }

                return QVariant( strValue );
            }

            else if ( key == "FontFamily" ) {
                if ( (strValue == "default") or (strValue == "sans-serif") ) {
                    strValue = QFontDatabase::systemFont( QFontDatabase::GeneralFont ).family();
                }

                else if ( (strValue == "fixed") or (strValue == "monospace") ) {
                    strValue = QFontDatabase::systemFont( QFontDatabase::FixedFont ).family();
                }

                return QVariant( strValue );
            }

            /** We need to return the QFont object */
            else if ( key == "Font" ) {
                if ( (strValue == "default") or (strValue == "sans-serif") ) {
                    value = QFontDatabase::systemFont( QFontDatabase::GeneralFont );
                }

                else if ( (strValue == "fixed") or (strValue == "monospace") ) {
                    value = QFontDatabase::systemFont( QFontDatabase::FixedFont );
                }
            }
        }
    }

    return value;
}


void DFL::SettingsImpl::setValue( const QString& key, QVariant value ) {
    usrSett->setValue( key, value );
    usrSett->sync();
}


QStringList DFL::SettingsImpl::allKeys() {
    return map.keys();
}


bool DFL::SettingsImpl::hasKey( QString key ) {
    return map.contains( key );
}


ChangedSettings DFL::SettingsImpl::changedSettings() {
    /** Sync the QSettings */
    defSett->sync();
    sysSett->sync();
    usrSett->sync();

    SettingsMap     newMap = readCurrentSettings();
    ChangedSettings chSettList;

    /**
     * Check the settigns that have changed.
     * If a key present in @newMap is absent in @map, then it was newly added.
     * If the key is present in both maps, check if it was modified.
     */
    for ( QString key: newMap.keys() ) {
        /** Ignore session changes */
        if ( key.startsWith( "Session/" ) ) {
            continue;
        }

        /** Ignore changes to 'Core/' */
        if ( key.startsWith( "Core/" ) ) {
            continue;
        }

        /**
         * A new key was added to default settings.
         * This probably means that the app was updated.
         * User is strongly advised to restart the app.
         */
        if ( not map.keys().contains( key ) ) {
            ChangedSetting s{
                .key   = key,
                .value = QVariant(),
                .state = 'a',
            };

            chSettList << s;
            map[ key ] = newMap[ key ];
        }

        /**
         * Okay this key exists.
         * If the value has changed, update chSettList.
         * Also update @map.
         */
        else {
            if ( newMap[ key ] != map[ key ] ) {
                ChangedSetting s{
                    .key   = key,
                    .value = newMap[ key ],
                    .state = 'm',
                };

                chSettList << s;

                /**
                 * Update the value only if it's valid.
                 * This prevents null-values keys in the middle of a running app.
                 */
                if ( newMap[ key ].isValid() ) {
                    map[ key ] = newMap[ key ];
                }
            }
        }
    }

    /** Check if some key was removed. See if all keys present in @map exist in @newMap */
    for ( QString key: map.keys() ) {
        /** Ignore session changes */
        if ( key.startsWith( "Session/" ) ) {
            continue;
        }

        /** Ignore changes to 'Core/' */
        if ( key.startsWith( "Core/" ) ) {
            continue;
        }

        /**
         * A key was removed from the default settings.
         * This probably means that the app was updated.
         * User is strongly advised to restart the app.
         */
        if ( not newMap.keys().contains( key ) ) {
            ChangedSetting s{
                .key   = key,
                .value = QVariant(),
                .state = 'd',
            };

            chSettList << s;
        }
    }

    return chSettList;
}


SettingsMap DFL::SettingsImpl::readCurrentSettings() {
    /** Sync the QSettings */
    defSett->sync();
    sysSett->sync();
    usrSett->sync();

    /** Simply read all the settings as-is and store them in curMap */
    SettingsMap curMap;

    /** These are non 'Session' settings */
    for ( QString key: defSett->allKeys() ) {
        QVariant value;

        if ( usrSett->contains( key ) ) {
            value = usrSett->value( key );
        }

        else if ( sysSett->contains( key ) ) {
            value = sysSett->value( key );
        }

        else {
            value = defSett->value( key );
        }

        curMap[ key ] = value;
    }

    /** Load the 'Session' settings */
    usrSett->beginGroup( "Session" );
    for ( QString key: usrSett->childKeys() ) {
        curMap[ "Session/" + key ] = usrSett->value( key );
    }
    usrSett->endGroup();

    /**
     * Read dynamic settings
     * Dynamic settings will under groups containing ':'
     * For example: [Output:eDP-1]
     */
    for ( QString group : usrSett->childGroups() ) {
        if ( group.contains( ":" ) ) {
            usrSett->beginGroup( group );
            for ( QString key: usrSett->childKeys() ) {
                curMap[ group + "/" + key ] = usrSett->value( key );
            }
            usrSett->endGroup();
        }
    }

    /** Return curMap */
    return curMap;
}
